#!/usr/bin/env python
# -*- coding: utf-8 -*-


class TrackedObj(object):
    def __init__(self, val):
        self.val = val

    def __str__(self):
        return '[%s]' % self.val


class TrackField(TrackedObj):
    pass


class TrackIndex(TrackedObj):
    pass


class TrackVariant(TrackedObj):
    pass


class Tracker(object):
    def __init__(self):
        self.cur = []

    def push(self, obj):
        self.cur.append(obj)

    def push_field(self, obj):
        self.push(TrackField(obj))

    def push_index(self, obj):
        self.push(TrackIndex(obj))

    def push_variant(self, obj):
        self.push(TrackVariant(obj))

    def pop(self):
        self.cur.pop()

    def __str__(self):
        return ''.join([str(x) for x in self.cur])


class ArchiveException(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.tracker = kwargs.get('tracker', None)



from .base_types import XmrType
from .obj_helper import eq_obj_contents, is_type, slot_obj_dict


class BlobType(XmrType):
    """
    Binary data

    Represented as bytearray() or a list of values in data structures.
    Not wrapped in the BlobType, the BlobType is only a scheme descriptor.
    Behaves in the same way as primitive types

    Supports also the wrapped version (__init__, DATA_ATTR, eq, repr...),
    """

    DATA_ATTR = "data"
    FIX_SIZE = 0
    SIZE = 0

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise ValueError()
        if len(args) > 0:
            setattr(self, self.DATA_ATTR, args[0])
        if 'SIZE' in kwargs:
            self.SIZE = kwargs['SIZE']
        if 'FIX_SIZE' in kwargs:
            self.FIX_SIZE = kwargs['FIX_SIZE']

    def __eq__(self, rhs):
        return eq_obj_contents(self, rhs)

    def __repr__(self):
        dct = slot_obj_dict(self) if hasattr(self, "__slots__") else self.__dict__
        return "<%s: %s>" % (self.__class__.__name__, dct)


class UnicodeType(XmrType):
    pass


class VariantType(XmrType):
    """
    Union of types, variant tags needed. is only one of the types. List in typedef, enum.
    Wraps the variant type in order to unambiguously support variant of variants.
    Supports also unwrapped value using type system to distinguish variants - simplifies the construction.
    """
    MFIELDS = []
    WRAPS_VALUE = False

    def __init__(self, *args, **kwargs):
        self.variant_elem = None
        self.variant_elem_type = None

        fname, fval = None, None
        if len(args) > 0:
            fname, fval = self.find_fdef(self.f_specs(), args[0])[0], args[0]
        if len(kwargs) > 0:
            key = list(kwargs.keys())[0]
            fname, fval = key, kwargs[key]
        if fname:
            self.set_variant(fname, fval)

    @classmethod
    def f_specs(cls):
        return cls.MFIELDS

    @staticmethod
    def find_fdef(fields, elem):
        for x in fields:
            if isinstance(elem, x[1]):
                return x
        raise ValueError('Unrecognized variant')

    def set_variant(self, fname, fvalue):
        self.variant_elem = fname
        self.variant_elem_type = fvalue.__class__
        setattr(self, fname, fvalue)

    def __eq__(self, rhs):
        return eq_obj_contents(self, rhs)

    def __repr__(self):
        dct = slot_obj_dict(self) if hasattr(self, "__slots__") else self.__dict__
        return "<%s: %s>" % (self.__class__.__name__, dct)


class ContainerType(XmrType):
    """
    Array of elements
    Represented as a real array in the data structures, not wrapped in the ContainerType.
    The Container type is used only as a schema descriptor for serialization.
    """

    FIX_SIZE = 0
    SIZE = 0
    ELEM_TYPE = None


class TupleType(XmrType):
    MFIELDS = []  # simple types without file name

    @classmethod
    def f_specs(cls):
        return cls.MFIELDS


class MessageType(XmrType):
    MFIELDS = []

    def __init__(self, **kwargs):
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])

    def __eq__(self, rhs):
        return eq_obj_contents(self, rhs)

    def __repr__(self):
        dct = slot_obj_dict(self) if hasattr(self, "__slots__") else self.__dict__
        return "<%s: %s>" % (self.__class__.__name__, dct)

    @classmethod
    def f_specs(cls):
        return cls.MFIELDS

    def _field(self, fname=None, idx=None):
        fld = None
        specs = self.f_specs()
        if fname is not None:
            fld = [x for x in specs if x[0] == fname][0]
        elif idx is not None:
            fld = specs[idx]
        return fld

    async def _msg_field(self, ar, fname=None, idx=None, **kwargs):
        return await ar.message_field(self, self._field(fname=fname, idx=idx), **kwargs)


def container_elem_type(container_type, params):
    """
    Returns container element type

    :param container_type:
    :param params:
    :return:
    """
    elem_type = params[0] if params else None
    if elem_type is None:
        elem_type = container_type.ELEM_TYPE
    return elem_type


def gen_elem_array(size, elem_type=None):
    """
    Generates element array of given size and initializes with given type.
    Supports container type, used for pre-allocation before deserialization.
    :param size:
    :param elem_type:
    :return:
    """
    if elem_type is None or not callable(elem_type):
        return [elem_type] * size
    if is_type(elem_type, ContainerType):

        def elem_type():
            return []

    res = []
    for _ in range(size):
        res.append(elem_type())
    return res



from apps.monero.xmr.serialize.message_types import ContainerType, MessageType
from apps.monero.xmr.serialize_messages.base import ECKey


class EcdhTuple(MessageType):
    __slots__ = ("mask", "amount")

    @staticmethod
    def f_specs():
        return (("mask", ECKey), ("amount", ECKey))


class EcdhInfo(ContainerType):
    ELEM_TYPE = EcdhTuple



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
XMR types
"""

from micropython import const

from apps.monero.xmr.serialize.message_types import BlobType

_c0 = const(0)
_c1 = const(1)
_c32 = const(32)
_c64 = const(64)

#
# cryptonote_basic.h
#


class Hash(BlobType):
    __slots__ = ("data",)
    DATA_ATTR = "data"
    FIX_SIZE = _c1
    SIZE = _c32


class ECKey(BlobType):
    __slots__ = ("bytes",)
    DATA_ATTR = "bytes"
    FIX_SIZE = _c1
    SIZE = _c32


ECPoint = Hash
SecretKey = ECKey
ECPublicKey = ECPoint
KeyImage = ECPoint
KeyDerivation = ECPoint



from micropython import const

from apps.monero.xmr.serialize.message_types import ContainerType, MessageType
from apps.monero.xmr.serialize_messages.base import ECKey

_c0 = const(0)
_c1 = const(1)
_c32 = const(32)
_c64 = const(64)


class Key64(ContainerType):
    FIX_SIZE = _c1
    SIZE = _c64
    ELEM_TYPE = ECKey


class KeyV(ContainerType):
    FIX_SIZE = _c0
    ELEM_TYPE = ECKey


class KeyM(ContainerType):
    FIX_SIZE = _c0
    ELEM_TYPE = KeyV


class KeyVFix(ContainerType):
    FIX_SIZE = _c1
    ELEM_TYPE = ECKey


class KeyMFix(ContainerType):
    FIX_SIZE = _c1
    ELEM_TYPE = KeyVFix


class CtKey(MessageType):
    __slots__ = ("dest", "mask")

    @staticmethod
    def f_specs():
        return (("dest", ECKey), ("mask", ECKey))


class CtkeyV(ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = CtKey


class CtkeyM(ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = CtkeyV




from apps.monero.xmr.serialize.base_types import UInt8, UVarintType
from apps.monero.xmr.serialize.erefs import eref
from apps.monero.xmr.serialize.message_types import ContainerType, MessageType
from apps.monero.xmr.serialize_messages.base import ECKey
from apps.monero.xmr.serialize_messages.ct_keys import CtKey, CtkeyM, CtkeyV, KeyM, KeyV
from apps.monero.xmr.serialize_messages.tx_ecdh import EcdhInfo, EcdhTuple
from apps.monero.xmr.serialize_messages.tx_prefix import TransactionPrefix, TxinToKey
from apps.monero.xmr.serialize_messages.tx_rsig import RctType
from apps.monero.xmr.serialize_messages.tx_rsig_boro import RangeSig
from apps.monero.xmr.serialize_messages.tx_rsig_bulletproof import Bulletproof
from apps.monero.xmr.serialize_messages.tx_sig import (
    Signature,
    SignatureArray,
    get_signature_size,
)


class MgSig(MessageType):
    __slots__ = ("ss", "cc", "II")

    @staticmethod
    def f_specs():
        return (("ss", KeyM), ("cc", ECKey))


class RctSigBase(MessageType):
    __slots__ = (
        "type",
        "txnFee",
        "message",
        "mixRing",
        "pseudoOuts",
        "ecdhInfo",
        "outPk",
    )

    @staticmethod
    def f_specs():
        return (
            ("type", UInt8),
            ("txnFee", UVarintType),
            ("message", ECKey),
            ("mixRing", CtkeyM),
            ("pseudoOuts", KeyV),
            ("ecdhInfo", EcdhInfo),
            ("outPk", CtkeyV),
        )

    async def serialize_rctsig_base(self, ar, inputs, outputs):
        """
        Custom serialization
        :param ar:
        :type ar: x.Archive
        :return:
        """
        await self._msg_field(ar, idx=0)
        if self.type == RctType.Null:
            return
        if (
            self.type != RctType.Full
            and self.type != RctType.FullBulletproof
            and self.type != RctType.Simple
            and self.type != RctType.SimpleBulletproof
        ):
            raise ValueError("Unknown type")

        await self._msg_field(ar, idx=1)
        if self.type == RctType.Simple:
            await ar.prepare_container(inputs, eref(self, "pseudoOuts"), KeyV)
            if ar.writing and len(self.pseudoOuts) != inputs:
                raise ValueError("pseudoOuts size mismatch")

            for i in range(inputs):
                await ar.field(eref(self.pseudoOuts, i), KeyV.ELEM_TYPE)

        await ar.prepare_container(outputs, eref(self, "ecdhInfo"), EcdhTuple)
        if ar.writing and len(self.ecdhInfo) != outputs:
            raise ValueError("EcdhInfo size mismatch")

        for i in range(outputs):
            await ar.field(eref(self.ecdhInfo, i), EcdhInfo.ELEM_TYPE)

        await ar.prepare_container((outputs), eref(self, "outPk"), CtKey)
        if ar.writing and len(self.outPk) != outputs:
            raise ValueError("outPk size mismatch")

        for i in range(outputs):
            await ar.field(eref(self.outPk[i], "mask"), ECKey)


class RctSigPrunable(MessageType):
    __slots__ = ("rangeSigs", "bulletproofs", "MGs", "pseudoOuts")

    @staticmethod
    def f_specs():
        return (
            ("rangeSigs", ContainerType, RangeSig),
            ("bulletproofs", ContainerType, Bulletproof),
            ("MGs", ContainerType, MgSig),
            ("pseudoOuts", KeyV),
        )

    async def serialize_rctsig_prunable(self, ar, type, inputs, outputs, mixin):
        """
        Serialize rct sig
        :param ar:
        :type ar: x.Archive
        :param type:
        :param inputs:
        :param outputs:
        :param mixin:
        :return:
        """
        if type == RctType.Null:
            return True

        if (
            type != RctType.Full
            and type != RctType.FullBulletproof
            and type != RctType.Simple
            and type != RctType.SimpleBulletproof
        ):
            raise ValueError("Unknown type")

        if type == RctType.SimpleBulletproof or type == RctType.FullBulletproof:
            if len(self.bulletproofs) != outputs:
                raise ValueError("Bulletproofs size mismatch")

            await ar.prepare_container(
                outputs, eref(self, "bulletproofs"), elem_type=Bulletproof
            )
            for i in range(len(self.bulletproofs)):
                await ar.field(elem=eref(self.bulletproofs, i), elem_type=Bulletproof)

        else:
            await ar.prepare_container(
                outputs, eref(self, "rangeSigs"), elem_type=RangeSig
            )
            if len(self.rangeSigs) != outputs:
                raise ValueError("rangeSigs size mismatch")

            for i in range(len(self.rangeSigs)):
                await ar.field(elem=eref(self.rangeSigs, i), elem_type=RangeSig)

        # We keep a byte for size of MGs, because we don't know whether this is
        # a simple or full rct signature, and it's starting to annoy the hell out of me
        mg_elements = (
            inputs if type == RctType.Simple or type == RctType.SimpleBulletproof else 1
        )
        await ar.prepare_container(mg_elements, eref(self, "MGs"), elem_type=MgSig)
        if len(self.MGs) != mg_elements:
            raise ValueError("MGs size mismatch")

        for i in range(mg_elements):
            # We save the MGs contents directly, because we want it to save its
            # arrays and matrices without the size prefixes, and the load can't
            # know what size to expect if it's not in the data

            await ar.prepare_container(
                mixin + 1, eref(self.MGs[i], "ss"), elem_type=KeyM
            )
            if ar.writing and len(self.MGs[i].ss) != mixin + 1:
                raise ValueError("MGs size mismatch")

            for j in range(mixin + 1):
                mg_ss2_elements = 1 + (
                    1
                    if type == RctType.Simple or type == RctType.SimpleBulletproof
                    else inputs
                )
                await ar.prepare_container(
                    mg_ss2_elements, eref(self.MGs[i].ss, j), elem_type=KeyM.ELEM_TYPE
                )

                if ar.writing and len(self.MGs[i].ss[j]) != mg_ss2_elements:
                    raise ValueError("MGs size mismatch 2")

                for k in range(mg_ss2_elements):
                    await ar.field(eref(self.MGs[i].ss[j], k), elem_type=KeyV.ELEM_TYPE)

            await ar.field(eref(self.MGs[i], "cc"), elem_type=ECKey)

        if type == RctType.SimpleBulletproof:
            await ar.prepare_container(inputs, eref(self, "pseudoOuts"), elem_type=KeyV)
            if ar.writing and len(self.pseudoOuts) != inputs:
                raise ValueError("pseudoOuts size mismatch")

            for i in range(inputs):
                await ar.field(eref(self.pseudoOuts, i), elem_type=KeyV.ELEM_TYPE)


class RctSig(RctSigBase):
    @staticmethod
    def f_specs():
        return RctSigBase.f_specs() + (("p", RctSigPrunable),)


class Transaction(TransactionPrefix):
    @staticmethod
    def f_specs():
        return TransactionPrefix.f_specs() + (
            ("signatures", ContainerType, SignatureArray),
            ("rct_signatures", RctSig),
        )

    async def serialize_archive(self, ar):
        """
        Serialize the transaction
        :param ar:
        :type ar: x.Archive
        :return:
        """
        # Transaction prefix serialization first.
        await ar.message(self, TransactionPrefix)

        if self.version == 1:
            await ar.prepare_container(
                len(self.vin), eref(self, "signatures"), elem_type=SignatureArray
            )
            signatures_not_expected = len(self.signatures) == 0
            if not signatures_not_expected and len(self.vin) != len(self.signatures):
                raise ValueError("Signature size mismatch")

            for i in range(len(self.vin)):
                sig_size = get_signature_size(self.vin[i])
                if signatures_not_expected:
                    if 0 == sig_size:
                        continue
                    else:
                        raise ValueError("Unexpected sig")

                await ar.prepare_container(
                    sig_size, eref(self.signatures, i), elem_type=Signature
                )
                if sig_size != len(self.signatures[i]):
                    raise ValueError("Unexpected sig size")

                await ar.message(self.signatures[i], Signature)

        else:
            if len(self.vin) == 0:
                return

            await ar.prepare_message(eref(self, "rct_signatures"), RctSig)
            await self.rct_signatures.serialize_rctsig_base(
                ar, len(self.vin), len(self.vout)
            )

            if self.rct_signatures.type != RctType.Null:
                mixin_size = (
                    len(self.vin[0].key_offsets) - 1
                    if len(self.vin) > 0 and isinstance(self.vin[0], TxinToKey)
                    else 0
                )
                await ar.prepare_message(eref(self.rct_signatures, "p"), RctSigPrunable)
                await self.rct_signatures.p.serialize_rctsig_prunable(
                    ar,
                    self.rct_signatures.type,
                    len(self.vin),
                    len(self.vout),
                    mixin_size,
                )
        return self



from micropython import const

from apps.monero.xmr.serialize.base_types import UInt8, UVarintType
from apps.monero.xmr.serialize.message_types import (
    BlobType,
    ContainerType,
    MessageType,
    VariantType,
)
from apps.monero.xmr.serialize_messages.base import ECPublicKey, Hash, KeyImage

_c0 = const(0)
_c1 = const(1)
_c32 = const(32)
_c64 = const(64)


class TxoutToScript(MessageType):
    __slots__ = ("keys", "script")
    VARIANT_CODE = 0x0

    @staticmethod
    def f_specs():
        return (("keys", ContainerType, ECPublicKey), ("script", ContainerType, UInt8))


class TxoutToKey(MessageType):
    __slots__ = ("key",)
    VARIANT_CODE = 0x2

    @staticmethod
    def f_specs():
        return (("key", ECPublicKey),)


class TxoutToScriptHash(MessageType):
    __slots__ = ("hash",)
    VARIANT_CODE = 0x1

    @staticmethod
    def f_specs():
        return (("hash", Hash),)


class TxoutTargetV(VariantType):
    @staticmethod
    def f_specs():
        return (
            ("txout_to_script", TxoutToScript),
            ("txout_to_scripthash", TxoutToScriptHash),
            ("txout_to_key", TxoutToKey),
        )


class TxinGen(MessageType):
    __slots__ = ("height",)
    VARIANT_CODE = 0xff

    @staticmethod
    def f_specs():
        return (("height", UVarintType),)


class TxinToKey(MessageType):
    __slots__ = ("amount", "key_offsets", "k_image")
    VARIANT_CODE = 0x2

    @staticmethod
    def f_specs():
        return (
            ("amount", UVarintType),
            ("key_offsets", ContainerType, UVarintType),
            ("k_image", KeyImage),
        )


class TxinToScript(MessageType):
    __slots__ = ()
    VARIANT_CODE = _c0


class TxinToScriptHash(MessageType):
    __slots__ = ()
    VARIANT_CODE = _c1


class TxInV(VariantType):
    @staticmethod
    def f_specs():
        return (
            ("txin_gen", TxinGen),
            ("txin_to_script", TxinToScript),
            ("txin_to_scripthash", TxinToScriptHash),
            ("txin_to_key", TxinToKey),
        )


class TxOut(MessageType):
    __slots__ = ("amount", "target")

    @staticmethod
    def f_specs():
        return (("amount", UVarintType), ("target", TxoutTargetV))


class TransactionPrefix(MessageType):
    @staticmethod
    def f_specs():
        return (
            ("version", UVarintType),
            ("unlock_time", UVarintType),
            ("vin", ContainerType, TxInV),
            ("vout", ContainerType, TxOut),
            ("extra", ContainerType, UInt8),
        )


class TransactionPrefixExtraBlob(TransactionPrefix):
    # noinspection PyTypeChecker
    @staticmethod
    def f_specs():
        return TransactionPrefix.f_specs()[:-1] + (("extra", BlobType),)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
XMR types
'''


from . import xmrserialize as x
from . import xmrrpc
from .xmrserialize import eref


#
# cryptonote_basic.h
#


class Hash(x.BlobType):
    __slots__ = ['data']
    DATA_ATTR = 'data'
    FIX_SIZE = 1
    SIZE = 32


class ECKey(x.BlobType):
    __slots__ = ['bytes']
    DATA_ATTR = 'bytes'
    FIX_SIZE = 1
    SIZE = 32


class ECPoint(x.BlobType):
    __slots__ = ['data']
    DATA_ATTR = 'data'
    FIX_SIZE = 1
    SIZE = 32


class SecretKey(ECKey):
    __slots__ = ['bytes']


class ECPublicKey(ECPoint):
    __slots__ = ['data']


class KeyImage(ECPoint):
    __slots__ = ['data']


class KeyDerivation(ECPoint):
    __slots__ = ['data']


class UnorderedSet(x.ContainerType):
    pass


class TxoutToScript(x.MessageType):
    __slots__ = ['keys', 'script']
    VARIANT_CODE = 0x0
    MFIELDS = [
        ('keys', x.ContainerType, ECPublicKey),
        ('script', x.ContainerType, x.UInt8),
    ]


class TxoutToKey(x.MessageType):
    __slots__ = ['key']
    VARIANT_CODE = 0x2
    MFIELDS = [
        ('key', ECPublicKey),
    ]


class TxoutToScriptHash(x.MessageType):
    __slots__ = ['hash']
    VARIANT_CODE = 0x1
    MFIELDS = [
        ('hash', Hash),
    ]


class TxoutTargetV(x.VariantType):
    MFIELDS = [
        ('txout_to_script', TxoutToScript),
        ('txout_to_scripthash', TxoutToScriptHash),
        ('txout_to_key', TxoutToKey),
    ]


class TxinGen(x.MessageType):
    __slots__ = ['height']
    VARIANT_CODE = 0xff
    BOOST_VARIANT_CODE = 0x0
    MFIELDS = [
        ('height', x.UVarintType),
    ]


class TxinToKey(x.MessageType):
    __slots__ = ['amount', 'key_offsets', 'k_image']
    VARIANT_CODE = 0x2
    BOOST_VARIANT_CODE = 0x3
    MFIELDS = [
        ('amount', x.UVarintType),
        ('key_offsets', x.ContainerType, x.UVarintType),
        ('k_image', KeyImage),
    ]


class TxinToScript(x.MessageType):
    __slots__ = []
    VARIANT_CODE = 0x0
    BOOST_VARIANT_CODE = 0x1
    MFIELDS = []


class TxinToScriptHash(x.MessageType):
    __slots__ = []
    VARIANT_CODE = 0x1
    BOOST_VARIANT_CODE = 0x2
    MFIELDS = []


class TxInV(x.VariantType):
    MFIELDS = [
        ('txin_gen', TxinGen),
        ('txin_to_script', TxinToScript),
        ('txin_to_scripthash', TxinToScriptHash),
        ('txin_to_key', TxinToKey),
    ]


class TxOut(x.MessageType):
    __slots__ = ['amount', 'target']
    MFIELDS = [
        ('amount', x.UVarintType),
        ('target', TxoutTargetV),
    ]


class TransactionPrefix(x.MessageType):
    MFIELDS = [
        ('version', x.UVarintType),
        ('unlock_time', x.UVarintType),
        ('vin', x.ContainerType, TxInV),
        ('vout', x.ContainerType, TxOut),
        ('extra', x.ContainerType, x.UInt8),
    ]


class TransactionPrefixExtraBlob(TransactionPrefix):
    # noinspection PyTypeChecker
    MFIELDS = TransactionPrefix.MFIELDS[:-1] + [
        ('extra', x.BlobType),
    ]


#
# rctTypes.h
#


class Key64(x.ContainerType):
    FIX_SIZE = 1
    SIZE = 64
    BOOST_RAW_ARRAY = True
    ELEM_TYPE = ECKey


class KeyV(x.ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = ECKey


class KeyM(x.ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = KeyV


class KeyVFix(x.ContainerType):
    FIX_SIZE = 1
    ELEM_TYPE = ECKey


class KeyMFix(x.ContainerType):
    FIX_SIZE = 1
    ELEM_TYPE = KeyVFix


class CtKey(x.MessageType):
    __slots__ = ['dest', 'mask']
    MFIELDS = [
        ('dest', ECKey),
        ('mask', ECKey),
    ]


class CtkeyV(x.ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = CtKey


class CtkeyM(x.ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = CtkeyV


class MultisigKLRki(x.MessageType):
    MFIELDS = [
        ('K', ECKey),
        ('L', ECKey),
        ('R', ECKey),
        ('ki', ECKey),
    ]


class MultisigOut(x.MessageType):
    MFIELDS = [
        ('c', x.ContainerType, ECKey),
    ]


class EcdhTuple(x.MessageType):
    __slots__ = ['mask', 'amount']
    MFIELDS = [
        ('mask', ECKey),
        ('amount', ECKey),
    ]


class BoroSig(x.MessageType):
    __slots__ = ['s0', 's1', 'ee']
    MFIELDS = [
        ('s0', Key64),
        ('s1', Key64),
        ('ee', ECKey),
    ]


class MgSig(x.MessageType):
    __slots__ = ['ss', 'cc', 'II']
    MFIELDS = [
        ('ss', KeyM),
        ('cc', ECKey),
    ]


class RangeSig(x.MessageType):
    __slots__ = ['asig', 'Ci']
    MFIELDS = [
        ('asig', BoroSig),
        ('Ci', Key64),
    ]


class Bulletproof(x.MessageType):
    __slots__ = ['V', 'A', 'S', 'T1', 'T2', 'taux', 'mu', 'L', 'R', 'a', 'b', 't']
    MFIELDS = [
        ('A', ECKey),
        ('S', ECKey),
        ('T1', ECKey),
        ('T2', ECKey),
        ('taux', ECKey),
        ('mu', ECKey),
        ('L', KeyV),
        ('R', KeyV),
        ('a', ECKey),
        ('b', ECKey),
        ('t', ECKey),
    ]

    async def boost_serialize(self, ar, version=None):
        await ar.message_fields(self, [('V', ECKey)] + self.MFIELDS)
        return self


class EcdhInfo(x.ContainerType):
    ELEM_TYPE = EcdhTuple


async def boost_out_pk(ar, out_pk, version):
    if ar.writing:
        out_pk_m = [k.mask for k in x.get_elem(out_pk)]
        await ar.field(out_pk_m, KeyV)

    else:
        outs = await ar.field(None, KeyV)
        pks = [CtKey(dest=b'\x01'+(b'\x00'*31), mask=k) for k in outs]
        x.set_elem(out_pk, pks)
        return pks


class RctSigBase(x.MessageType):
    __slots__ = ['type', 'txnFee', 'message', 'mixRing', 'pseudoOuts', 'ecdhInfo', 'outPk']
    MFIELDS = [
        ('type', x.UInt8),
        ('txnFee', x.UVarintType),
        ('message', ECKey),
        ('mixRing', CtkeyM),
        ('pseudoOuts', KeyV),
        ('ecdhInfo', EcdhInfo),
        ('outPk', CtkeyV),
    ]

    async def serialize_rctsig_base(self, ar, inputs, outputs):
        """
        Custom serialization
        :param ar:
        :type ar: x.Archive
        :return:
        """
        await self._msg_field(ar, idx=0)
        if self.type == RctType.Null:
            return
        if self.type != RctType.Full and self.type != RctType.FullBulletproof and \
                self.type != RctType.Simple and self.type != RctType.SimpleBulletproof:
            raise ValueError('Unknown type')

        await self._msg_field(ar, idx=1)
        if self.type == RctType.Simple:
            await ar.tag('pseudoOuts')
            await ar.begin_array()
            await ar.prepare_container(inputs, eref(self, 'pseudoOuts'), KeyV)
            if ar.writing and len(self.pseudoOuts) != inputs:
                raise ValueError('pseudoOuts size mismatch')

            for i in range(inputs):
                await ar.field(eref(self.pseudoOuts, i), KeyV.ELEM_TYPE)
            await ar.end_array()

        await ar.tag('ecdhInfo')
        await ar.begin_array()
        await ar.prepare_container(outputs, eref(self, 'ecdhInfo'), EcdhTuple)
        if ar.writing and len(self.ecdhInfo) != outputs:
            raise ValueError('EcdhInfo size mismatch')

        for i in range(outputs):
            await ar.field(eref(self.ecdhInfo, i), EcdhInfo.ELEM_TYPE)
        await ar.end_array()

        await ar.tag('outPk')
        await ar.begin_array()
        await ar.prepare_container((outputs), eref(self, 'outPk'), CtKey)
        if ar.writing and len(self.outPk) != outputs:
            raise ValueError('outPk size mismatch')

        for i in range(outputs):
            await ar.field(eref(self.outPk[i], 'mask'), ECKey)
        await ar.end_array()

    async def boost_serialize(self, ar, version=None):
        await self._msg_field(ar, 'type')
        if self.type == RctType.Simple:
            await self._msg_field(ar, 'pseudoOuts')
        await self._msg_field(ar, 'ecdhInfo')
        await boost_out_pk(ar, eref(self, 'outPk'), version)
        await self._msg_field(ar, 'txnFee')
        return self


class RctType(object):
    Null = 0
    Full = 1
    Simple = 2
    Bulletproof = 3
    FullBulletproof = 3     # pre v9
    SimpleBulletproof = 4   # pre v9


class RctSigPrunable(x.MessageType):
    __slots__ = ['rangeSigs', 'bulletproofs', 'MGs', 'pseudoOuts']
    MFIELDS = [
        ('rangeSigs', x.ContainerType, RangeSig),
        ('bulletproofs', x.ContainerType, Bulletproof),
        ('MGs', x.ContainerType, MgSig),
        ('pseudoOuts', KeyV),
    ]

    async def serialize_rctsig_prunable(self, ar, type, inputs, outputs, mixin):
        """
        Serialize rct sig
        :param ar:
        :type ar: x.Archive
        :param type:
        :param inputs:
        :param outputs:
        :param mixin:
        :return:
        """
        if type == RctType.Null:
            return True

        if type != RctType.Full and type != RctType.FullBulletproof and \
                type != RctType.Simple and type != RctType.SimpleBulletproof:
            raise ValueError('Unknown type')

        if type == RctType.SimpleBulletproof or type == RctType.FullBulletproof:
            await ar.tag('bp')
            await ar.begin_array()
            bps = [0]
            if ar.writing:
                bps[0] = len(self.bulletproofs)
            await ar.field(elem=eref(bps, 0), elem_type=x.UVarintType)
            await ar.prepare_container(bps[0], eref(self, 'bulletproofs'), elem_type=Bulletproof)

            for i in range(bps[0]):
                await ar.field(elem=eref(self.bulletproofs, i), elem_type=Bulletproof)
            await ar.end_array()

        else:
            await ar.tag('rangeSigs')
            await ar.begin_array()
            await ar.prepare_container(outputs, eref(self, 'rangeSigs'), elem_type=RangeSig)
            if len(self.rangeSigs) != outputs:
                raise ValueError('rangeSigs size mismatch')

            for i in range(len(self.rangeSigs)):
                await ar.field(elem=eref(self.rangeSigs, i), elem_type=RangeSig)
            await ar.end_array()

        await ar.tag('MGs')
        await ar.begin_array()

        # We keep a byte for size of MGs, because we don't know whether this is
        # a simple or full rct signature, and it's starting to annoy the hell out of me
        is_full = type == RctType.Full
        mg_elements = inputs if not is_full else 1
        await ar.prepare_container(mg_elements, eref(self, 'MGs'), elem_type=MgSig)
        if len(self.MGs) != mg_elements:
            raise ValueError('MGs size mismatch')

        for i in range(mg_elements):
            # We save the MGs contents directly, because we want it to save its
            # arrays and matrices without the size prefixes, and the load can't
            # know what size to expect if it's not in the data
            await ar.begin_object()
            await ar.tag('ss')
            await ar.begin_array()

            await ar.prepare_container(mixin + 1, eref(self.MGs[i], 'ss'), elem_type=KeyM)
            if ar.writing and len(self.MGs[i].ss) != mixin + 1:
                raise ValueError('MGs size mismatch')

            for j in range(mixin + 1):
                await ar.begin_array()
                mg_ss2_elements = 1 + (1 if not is_full else inputs)
                await ar.prepare_container(mg_ss2_elements, eref(self.MGs[i].ss, j), elem_type=KeyM.ELEM_TYPE)

                if ar.writing and len(self.MGs[i].ss[j]) != mg_ss2_elements:
                    raise ValueError('MGs size mismatch 2')

                for k in range(mg_ss2_elements):
                    await ar.field(eref(self.MGs[i].ss[j], k), elem_type=KeyV.ELEM_TYPE)
                await ar.end_array()

            await ar.tag('cc')
            await ar.field(eref(self.MGs[i], 'cc'), elem_type=ECKey)
            await ar.end_object()
        await ar.end_array()

        if type in [RctType.FullBulletproof, RctType.SimpleBulletproof]:
            await ar.begin_array()
            await ar.prepare_container(inputs, eref(self, 'pseudoOuts'), elem_type=KeyV)
            if ar.writing and len(self.pseudoOuts) != inputs:
                raise ValueError('pseudoOuts size mismatch')

            for i in range(inputs):
                await ar.field(eref(self.pseudoOuts, i), elem_type=KeyV.ELEM_TYPE)
            await ar.end_array()

    async def boost_serialize(self, ar, version):
        await self._msg_field(ar, 'rangeSigs')
        if self.rangeSigs is None or len(self.rangeSigs) == 0:
            await self._msg_field(ar, 'bulletproofs')
        await self._msg_field(ar, 'MGs')
        if self.rangeSigs is None or len(self.rangeSigs) == 0:
            await self._msg_field(ar, 'pseudoOuts')
        return self


class RctSig(RctSigBase):
    # noinspection PyTypeChecker
    MFIELDS = RctSigBase.MFIELDS + [
        ('p', RctSigPrunable),
    ]


class Signature(x.MessageType):
    __slots__ = ['c', 'r']
    MFIELDS = [
        ('c', ECKey),
        ('r', ECKey),
    ]

    async def serialize_archive(self, ar):
        ar.field(eref(self, 'c'), ECKey)
        ar.field(eref(self, 'r'), ECKey)
        return self


class SignatureArray(x.ContainerType):
    FIX_SIZE = 0
    ELEM_TYPE = Signature


def get_signature_size(msg):
    """
    Returns a signature size for the input
    :param msg:
    :return:
    """
    if isinstance(msg, (TxinGen, TxinToScript, TxinToScriptHash)):
        return 0
    elif isinstance(msg, TxinToKey):
        return len(msg.key_offsets)
    else:
        raise ValueError('Unknown tx in')


class Transaction(TransactionPrefix):
    # noinspection PyTypeChecker
    MFIELDS = TransactionPrefix.MFIELDS + [
        ('signatures', x.ContainerType, SignatureArray),
        ('rct_signatures', RctSig),
    ]

    async def serialize_archive(self, ar):
        """
        Serialize the transaction
        :param ar:
        :type ar: x.Archive
        :return:
        """
        # Transaction prefix serialization first.
        await ar.message(self, TransactionPrefix)

        if self.version == 1:
            await ar.tag('signatures')
            await ar.begin_array()
            await ar.prepare_container(len(self.vin), eref(self, 'signatures'), elem_type=SignatureArray)
            signatures_not_expected = len(self.signatures) == 0
            if not signatures_not_expected and len(self.vin) != len(self.signatures):
                raise ValueError('Signature size mismatch')

            for i in range(len(self.vin)):
                sig_size = get_signature_size(self.vin[i])
                if signatures_not_expected:
                    if 0 == sig_size:
                        continue
                    else:
                        raise ValueError('Unexpected sig')

                await ar.prepare_container(sig_size, eref(self.signatures, i), elem_type=SignatureArray)
                if sig_size != len(self.signatures[i]):
                    raise ValueError('Unexpected sig size')

                for j in range(sig_size):
                    await ar.field(eref(self.signatures[i], j), Signature)

        else:
            await ar.tag('rct_signatures')
            if len(self.vin) == 0:
                return

            await ar.begin_object()
            await ar.prepare_message(eref(self, 'rct_signatures'), RctSig)
            await self.rct_signatures.serialize_rctsig_base(ar, len(self.vin), len(self.vout))
            await ar.end_object()

            if self.rct_signatures.type != RctType.Null:
                mixin_size = len(self.vin[0].key_offsets) - 1 if len(self.vin) > 0 and isinstance(self.vin[0], TxinToKey) else 0
                await ar.tag('rctsig_prunable')
                await ar.begin_object()
                await ar.prepare_message(eref(self.rct_signatures, 'p'), RctSigPrunable)
                await self.rct_signatures.p.serialize_rctsig_prunable(ar, self.rct_signatures.type,
                                                                      len(self.vin), len(self.vout),
                                                                      mixin_size)
                await ar.end_object()
        return self

    async def boost_serialize(self, ar, version):
        await ar.message(self, TransactionPrefix, use_version=version)

        if self.version == 1:
            raise ValueError('TxV1 not supported')

        else:
            await ar.prepare_message(eref(self, 'rct_signatures'), RctSigBase)
            await ar.message(self.rct_signatures, RctSigBase)
            if self.rct_signatures.type != RctType.Null:
                await ar.prepare_message(eref(self.rct_signatures, 'p'), RctSigPrunable)
                await ar.message(self.rct_signatures.p, RctSigPrunable)
        return self


class BlockHeader(x.MessageType):
    MFIELDS = [
        ('major_version', x.UVarintType),  # x.UInt8
        ('minor_version', x.UVarintType),  # x.UInt8
        ('timestamp', x.UVarintType),  # x.UInt64
        ('prev_id', Hash),
        ('nonce', x.UInt32),
    ]


class HashVector(x.ContainerType):
    ELEM_TYPE = Hash


class Block(BlockHeader):
    # noinspection PyTypeChecker
    MFIELDS = BlockHeader.MFIELDS + [
        ('miner_tx', Transaction),
        ('tx_hashes', HashVector),
    ]


class AccountPublicAddress(x.MessageType):
    __slots__ = ['m_spend_public_key', 'm_view_public_key']
    MFIELDS = [
        ('m_spend_public_key', ECPublicKey),
        ('m_view_public_key', ECPublicKey),
    ]


class SubaddressIndex(x.MessageType):
    __slots__ = ['major', 'minor']
    MFIELDS = [
        ('major', x.UInt32),
        ('minor', x.UInt32),
    ]


class MultisigLR(x.MessageType):
    __slots__ = ['L', 'R']
    BOOST_VERSION = 0
    MFIELDS = [
        ('L', ECKey),
        ('R', ECKey),
    ]


class MultisigInfo(x.MessageType):
    __slots__ = ['signer', 'LR', 'partial_key_images']
    BOOST_VERSION = 1
    MFIELDS = [
        ('signer', ECPublicKey),
        ('LR', x.ContainerType, MultisigLR),
        ('partial_key_images', x.ContainerType, KeyImage),
    ]


class MultisigStruct(x.MessageType):
    __slots__ = ['sigs', 'ignore', 'used_L', 'signing_keys', 'msout']
    MFIELDS = [
        ('sigs', RctSig),
        ('ignore', ECPublicKey),
        ('used_L', x.ContainerType, ECKey),
        ('signing_keys', x.ContainerType, ECPublicKey),
        ('msout', MultisigOut),
    ]


class TxExtraPadding(x.MessageType):
    __slots__ = ['size']
    TX_EXTRA_PADDING_MAX_COUNT = 255

    VARIANT_CODE = 0x0
    MFIELDS = [
        ('size', x.SizeT),
    ]

    async def serialize_archive(self, ar):
        if ar.writing:
            if self.size > self.TX_EXTRA_PADDING_MAX_COUNT:
                raise ValueError('Padding too big')
            for i in range(self.size):
                ar.uint(0, x.UInt8)

        else:
            self.size = 0
            buffer = bytearray(1)
            for i in range(self.TX_EXTRA_PADDING_MAX_COUNT+1):
                self.size += 1
                try:
                    nread = await ar.iobj.areadinto(buffer)
                    if nread == 0:
                        break
                except EOFError:
                    break

                if buffer[0] != 0:
                    raise ValueError('Padding error')
        return self


class TxExtraPubKey(x.MessageType):
    __slots__ = ['pub_key']
    VARIANT_CODE = 0x1
    MFIELDS = [
        ('pub_key', ECPublicKey),
    ]


class TxExtraNonce(x.MessageType):
    __slots__ = ['nonce']
    VARIANT_CODE = 0x2
    MFIELDS = [
        ('nonce', x.BlobType),
    ]


class TxExtraMergeMiningTag(x.MessageType):
    VARIANT_CODE = 0x3
    MFIELDS = [
        ('field_len', x.UVarintType),
        ('depth', x.UVarintType),
        ('merkle_root', Hash),
    ]


class TxExtraAdditionalPubKeys(x.MessageType):
    __slots__ = ['data']
    VARIANT_CODE = 0x4
    MFIELDS = [
        ('data', x.ContainerType, ECPublicKey),
    ]


class TxExtraMysteriousMinergate(x.MessageType):
    __slots__ = ['data']
    VARIANT_CODE = 0xde
    MFIELDS = [
        ('data', x.BlobType),
    ]


class TxExtraField(x.VariantType):
    MFIELDS = [
        ('tx_extra_padding', TxExtraPadding),
        ('tx_extra_pub_key', TxExtraPubKey),
        ('tx_extra_nonce', TxExtraNonce),
        ('tx_extra_merge_mining_tag', TxExtraMergeMiningTag),
        ('tx_extra_additional_pub_keys', TxExtraAdditionalPubKeys),
        ('tx_extra_mysterious_minergate', TxExtraMysteriousMinergate),
    ]


class TxExtraFields(x.ContainerType):
    ELEM_TYPE = TxExtraField


class OutputEntry(x.TupleType):
    MFIELDS = [
        x.UVarintType, CtKey  # original: x.UInt64
    ]


class TxSourceEntry(x.MessageType):
    BOOST_VERSION = 1
    MFIELDS = [
        ('outputs', x.ContainerType, OutputEntry),
        ('real_output', x.SizeT),
        ('real_out_tx_key', ECPublicKey),
        ('real_out_additional_tx_keys', x.ContainerType, ECPublicKey),
        ('real_output_in_tx_index', x.UInt64),
        ('amount', x.UInt64),
        ('rct', x.BoolType),
        ('mask', ECKey),
        ('multisig_kLRki', MultisigKLRki),
    ]

    async def boost_serialize(self, ar, version):
        if version < 1:
            raise ValueError('TxSourceEntry v1 supported only')
        await self._msg_field(ar, 'outputs')
        await self._msg_field(ar, 'real_output')
        await self._msg_field(ar, 'real_out_tx_key')
        await self._msg_field(ar, 'real_output_in_tx_index')
        await self._msg_field(ar, 'amount')
        await self._msg_field(ar, 'rct')
        await self._msg_field(ar, 'mask')
        await self._msg_field(ar, 'multisig_kLRki')
        await self._msg_field(ar, 'real_out_additional_tx_keys')
        return self


class TxDestinationEntry(x.MessageType):
    __slots__ = ['amount', 'addr', 'is_subaddress']
    BOOST_VERSION = 1
    MFIELDS = [
        ('amount', x.UVarintType),  # original: UInt64
        ('addr', AccountPublicAddress),
        ('is_subaddress', x.BoolType),
    ]


class TransferDetails(x.MessageType):
    BOOST_VERSION = 9
    MFIELDS = [
        ('m_block_height', x.UInt64),
        ('m_tx', TransactionPrefix),
        ('m_txid', Hash),
        ('m_internal_output_index', x.SizeT),
        ('m_global_output_index', x.UInt64),
        ('m_spent', x.BoolType),
        ('m_spent_height', x.UInt64),
        ('m_key_image', KeyImage),
        ('m_mask', ECKey),
        ('m_amount', x.UInt64),
        ('m_rct', x.BoolType),
        ('m_key_image_known', x.BoolType),
        ('m_pk_index', x.SizeT),
        ('m_subaddr_index', SubaddressIndex),
        ('m_key_image_partial', x.BoolType),
        ('m_multisig_k', x.ContainerType, ECKey),
        ('m_multisig_info', x.ContainerType, MultisigInfo),
    ]

    async def boost_serialize(self, ar, version):
        if version < 9:
            raise ValueError('TransferDetails v9 supported only')
        await self._msg_field(ar, 'm_block_height')
        await self._msg_field(ar, 'm_global_output_index')
        await self._msg_field(ar, 'm_internal_output_index')
        await self._msg_field(ar, 'm_tx')
        await self._msg_field(ar, 'm_spent')
        await self._msg_field(ar, 'm_key_image')
        await self._msg_field(ar, 'm_mask')
        await self._msg_field(ar, 'm_amount')
        await self._msg_field(ar, 'm_spent_height')
        await self._msg_field(ar, 'm_txid')
        await self._msg_field(ar, 'm_rct')
        await self._msg_field(ar, 'm_key_image_known')
        await self._msg_field(ar, 'm_pk_index')
        await self._msg_field(ar, 'm_subaddr_index')
        await self._msg_field(ar, 'm_multisig_info')
        await self._msg_field(ar, 'm_multisig_k')
        await self._msg_field(ar, 'm_key_image_partial')
        return self


class TxConstructionData(x.MessageType):
    BOOST_VERSION = 3
    MFIELDS = [
        ('sources', x.ContainerType, TxSourceEntry),
        ('change_dts', TxDestinationEntry),
        ('splitted_dsts', x.ContainerType, TxDestinationEntry),
        ('selected_transfers', x.ContainerType, x.SizeT),
        ('extra', x.ContainerType, x.UInt8),
        ('unlock_time', x.UInt64),
        ('use_rct', x.BoolType),
        # ('use_bulletproofs', x.BoolType),
        ('dests', x.ContainerType, TxDestinationEntry),
        ('subaddr_account', x.UInt32),
        ('subaddr_indices', x.ContainerType, x.UVarintType),  # original: x.UInt32
    ]

    async def boost_serialize(self, ar, version):
        if version < 2:
            raise ValueError('TXConstruction v2 supported only')
        await self._msg_field(ar, 'sources')
        await self._msg_field(ar, 'change_dts')
        await self._msg_field(ar, 'splitted_dsts')
        await self._msg_field(ar, 'extra')
        await self._msg_field(ar, 'unlock_time')
        await self._msg_field(ar, 'use_rct')
        await self._msg_field(ar, 'dests')
        await self._msg_field(ar, 'subaddr_account')
        await self._msg_field(ar, 'subaddr_indices')
        await self._msg_field(ar, 'selected_transfers')
        if version >= 3:
            await ar.message_field(self, ('use_bulletproofs', x.BoolType))
        return self


class PendingTransaction(x.MessageType):
    BOOST_VERSION = 3
    MFIELDS = [
        ('tx', Transaction),
        ('dust', x.UInt64),
        ('fee', x.UInt64),
        ('dust_added_to_fee', x.BoolType),
        ('change_dts', TxDestinationEntry),
        ('selected_transfers', x.ContainerType, x.SizeT),
        ('key_images', x.UnicodeType),
        ('tx_key', SecretKey),
        ('additional_tx_keys', x.ContainerType, SecretKey),
        ('dests', x.ContainerType, TxDestinationEntry),
        ('multisig_sigs', x.ContainerType, MultisigStruct),
        ('construction_data', TxConstructionData),
    ]

    async def boost_serialize(self, ar, version):
        if version < 3:
            raise ValueError('Pending transaction v3+ supported only')
        await self._msg_field(ar, idx=0)
        await self._msg_field(ar, idx=1)
        await self._msg_field(ar, idx=2)
        await self._msg_field(ar, idx=3)
        await self._msg_field(ar, idx=4)
        await self._msg_field(ar, 'key_images')
        await self._msg_field(ar, 'tx_key')
        await self._msg_field(ar, 'dests')
        await self._msg_field(ar, 'construction_data')
        await self._msg_field(ar, 'additional_tx_keys')
        await self._msg_field(ar, 'selected_transfers')
        await self._msg_field(ar, 'multisig_sigs')
        return self


class PendingTransactionVector(x.ContainerType):
    ELEM_TYPE = PendingTransaction


class UnsignedTxSet(x.MessageType):
    BOOST_VERSION = 0
    MFIELDS = [
        ('txes', x.ContainerType, TxConstructionData),
        ('transfers', x.ContainerType, TransferDetails),
    ]


class SignedTxSet(x.MessageType):
    BOOST_VERSION = 0
    MFIELDS = [
        ('ptx', PendingTransactionVector),
        ('key_images', x.ContainerType, KeyImage),
    ]


class MultisigTxSet(x.MessageType):
    BOOST_VERSION = 0
    MFIELDS = [
        ('m_ptx', PendingTransactionVector),
        ('m_signers', UnorderedSet, ECPublicKey),
    ]


class ChachaIv(x.BlobType):
    FIX_SIZE = 1
    SIZE = 8


class KeysFileData(x.MessageType):
    MFIELDS = [
        ('iv', ChachaIv),
        ('account_data', x.BlobType),
    ]


class CacheFileData(x.MessageType):
    MFIELDS = [
        ('iv', ChachaIv),
        ('cache_data', x.BlobType),
    ]


class AccountKeys(x.MessageType):
    MFIELDS = [
        ('m_account_address', AccountPublicAddress),
        ('m_spend_secret_key', SecretKey),
        ('m_view_secret_key', SecretKey),
        ('m_multisig_keys', x.ContainerType, SecretKey),
    ]

    async def kv_serialize(self, ar, obj=None, **kwargs):
        await self._msg_field(ar, 'm_account_address', obj=obj)
        await self._msg_field(ar, 'm_spend_secret_key', obj=obj)
        await self._msg_field(ar, 'm_view_secret_key', obj=obj)
        await ar.message_field(self, ('m_multisig_keys', xmrrpc.BlobFieldWrapper(x.ContainerType), SecretKey), obj=obj)
        return obj


class WalletKeyData(x.MessageType):
    MFIELDS = [
        ('m_creation_timestamp', x.UInt64),
        ('m_keys', AccountKeys),
    ]






from apps.monero.xmr.serialize.message_types import MessageType
from apps.monero.xmr.serialize_messages.base import ECKey
from apps.monero.xmr.serialize_messages.ct_keys import Key64


class BoroSig(MessageType):
    __slots__ = ("s0", "s1", "ee")

    @staticmethod
    def f_specs():
        return (("s0", Key64), ("s1", Key64), ("ee", ECKey))


class RangeSig(MessageType):
    __slots__ = ("asig", "Ci")

    @staticmethod
    def f_specs():
        return (("asig", BoroSig), ("Ci", Key64))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import binascii
import unittest
import json
import pkg_resources

import asyncio
import aiounittest

from .. import xmrserialize as x
from .. import xmrtypes as xmr
from .. import xmrobj as xmro
from .. import xmrjson as xmrjs
from .. import xmrrpc


__author__ = 'dusanklinec'


class XmrRpcTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(XmrRpcTest, self).__init__(*args, **kwargs)

    async def test_simple_msg(self):
        """
        TxinGen
        :return:
        """
        data = b'01110101010102010108146d5f6372656174696f6e5f74696d657374616d70057099935300000000066d5f6b6579730c0c116d5f6163636f756e745f616464726573730c08126d5f7370656e645f7075626c69635f6b65790a805a10cca900ee47a7f412cd661b29f5ab356d6a1951884593bb170b5ec8b6f2e8116d5f766965775f7075626c69635f6b65790a803b1da411527d062c9fedeb2dad669f2f5585a00a88462b8c95c809a630e5734c126d5f7370656e645f7365637265745f6b65790a80f2644a3dd97d43e87887e74d1691d52baa0614206ad1b0c239ff4aa3b501750a116d5f766965775f7365637265745f6b65790a804ce88c168e0f5f8d6524f712d5f8d7d83233b1e7a2a60b5aba5206cc0ea2bc08'
        data_bin = binascii.unhexlify(data)

        reader = x.MemoryReaderWriter(bytearray(data_bin))
        ar = xmrrpc.Archive(reader, False, modeled=True)

        section = {}
        await ar.root()
        await ar.section(section)
        self.assertIn('m_creation_timestamp', section)
        self.assertIn('m_keys', section)
        self.assertIn('m_account_address', section['m_keys'])
        self.assertIn('m_spend_public_key', section['m_keys']['m_account_address'])

        writer = x.MemoryReaderWriter()
        arw = xmrrpc.Archive(writer, True)
        await arw.root()
        await arw.section(section)

        reader2 = x.MemoryReaderWriter(bytearray(data_bin))
        ar2 = xmrrpc.Archive(reader2, False, modeled=False)

        section2 = {}
        await ar2.root()
        await ar2.section(section2)

        self.assertEqual(section['m_creation_timestamp'], section2['m_creation_timestamp'])
        self.assertDictEqual(section, section2)

    async def test_modeler(self):
        msg = xmr.AccountPublicAddress()
        msg.m_spend_public_key = b'\xff'*32
        msg.m_view_public_key = b'\xee'*32
        mdl = xmrrpc.Modeler(True)

        m2 = xmr.AccountKeys()
        m2.m_account_address = msg
        m2.m_spend_secret_key = b'\x12'*32
        m2.m_view_secret_key = b'\x15'*32
        m2.m_multisig_keys = [b'\x19'*32, b'\x22'*32]

        obj = await mdl.message(msg=m2)
        self.assertIsNotNone(obj)

        m2.m_multisig_keys = None
        obj = await mdl.message(msg=m2)
        self.assertIsNotNone(obj)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover






#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import base64
import unittest
import pkg_resources

import asyncio
import aiounittest

from .. import xmrserialize as x
from .. import xmrtypes as xmr


__author__ = 'dusanklinec'


class XmrSerializeBaseTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(XmrSerializeBaseTest, self).__init__(*args, **kwargs)

    async def test_varint(self):
        """
        Var int
        :return:
        """
        test_nums = [0, 1, 12, 44, 32, 63, 64, 127, 128, 255, 256, 1023, 1024, 8191, 8192,
                     2**16, 2**16 - 1, 2**32, 2**32 - 1, 2**64, 2**64 - 1, 2**72 - 1, 2**112]

        for test_num in test_nums:
            writer = x.MemoryReaderWriter()

            await x.dump_uvarint(writer, test_num)
            test_deser = await x.load_uvarint(x.MemoryReaderWriter(writer.get_buffer()))

            self.assertEqual(test_num, test_deser)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover





class XmrType:
    VERSION = 0


class UVarintType(XmrType):
    pass


class IntType(XmrType):
    WIDTH = 0
    SIGNED = 0
    VARIABLE = 0

    def __repr__(self):
        return "%s:<w: %s, sig: %s, var: %s>" % (
            self.__class__,
            self.WIDTH,
            self.SIGNED,
            self.VARIABLE,
        )


class BoolType(IntType):
    WIDTH = 1


class UInt8(IntType):
    WIDTH = 1


class Int8(IntType):
    SIGNED = 1
    WIDTH = 1


class UInt16(IntType):
    WIDTH = 2


class Int16(IntType):
    SIGNED = 1
    WIDTH = 2


class UInt32(IntType):
    WIDTH = 4


class Int32(IntType):
    SIGNED = 1
    WIDTH = 4


class UInt64(IntType):
    WIDTH = 8


class SizeT(UInt64):
    WIDTH = 8


class Int64(IntType):
    SIGNED = 1
    WIDTH = 8



class MemoryReaderWriter:
    def __init__(
        self,
        buffer=None,
        read_empty=False,
        threshold=None,
        do_gc=False,
        preallocate=None,
        gc=None,
        **kwargs
    ):
        self.buffer = buffer
        self.nread = 0
        self.nwritten = 0

        self.ndata = 0
        self.offset = 0
        self.woffset = 0

        self.read_empty = read_empty
        self.threshold = threshold
        self.do_gc = do_gc
        self.gc = gc

        if preallocate is not None:
            self.preallocate(preallocate)
        elif self.buffer is None:
            self.buffer = bytearray(0)
        else:
            self.woffset = len(buffer)

    def is_empty(self):
        return self.offset == len(self.buffer) or self.offset == self.woffset

    def preallocate(self, size):
        self.buffer = bytearray(size)
        self.offset = 0
        self.woffset = 0

    def readinto(self, buf):
        ln = len(buf)
        if not self.read_empty and ln > 0 and self.offset == len(self.buffer):
            raise EOFError

        nread = min(ln, len(self.buffer) - self.offset)
        for idx in range(nread):
            buf[idx] = self.buffer[self.offset + idx]

        self.offset += nread
        self.nread += nread
        self.ndata -= nread

        # Deallocation threshold triggered
        if self.threshold is not None and self.offset >= self.threshold:
            self.buffer = self.buffer[self.offset :]
            self.woffset -= self.offset
            self.offset = 0

            if self.do_gc:
                self.gc()

        return nread

    async def areadinto(self, buf):
        return self.readinto(buf)

    def write(self, buf):
        nwritten = len(buf)
        nall = len(self.buffer)
        towrite = nwritten
        bufoff = 0

        # Fill existing place in the buffer
        while towrite > 0 and nall - self.woffset > 0:
            self.buffer[self.woffset] = buf[bufoff]
            self.woffset += 1
            bufoff += 1
            towrite -= 1

        # Allocate next chunk if needed
        while towrite > 0:
            _towrite = min(32, towrite)
            chunk = bytearray(32)  # chunk size typical for EC point

            for i in range(_towrite):
                chunk[i] = buf[bufoff]
                self.woffset += 1
                bufoff += 1
                towrite -= 1

            self.buffer.extend(chunk)
            if self.do_gc:
                chunk = None  # dereference
                self.gc()

        self.nwritten += nwritten
        self.ndata += nwritten
        return nwritten

    async def awrite(self, buf):
        return self.write(buf)

    def get_buffer(self):
        mv = memoryview(self.buffer)
        return mv[self.offset : self.woffset]




from apps.monero.xmr.serialize.base_types import SizeT, UInt8, UVarintType
from apps.monero.xmr.serialize.message_types import (
    BlobType,
    ContainerType,
    MessageType,
    VariantType,
)
from apps.monero.xmr.serialize_messages.base import ECPublicKey, Hash


class TxExtraPadding(MessageType):
    __slots__ = ("size",)
    TX_EXTRA_PADDING_MAX_COUNT = 255

    VARIANT_CODE = 0x0

    @staticmethod
    def f_specs():
        return (("size", SizeT),)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size = 0

    async def serialize_archive(self, ar):
        if ar.writing:
            if self.size > self.TX_EXTRA_PADDING_MAX_COUNT:
                raise ValueError("Padding too big")
            for i in range(self.size):
                ar.uint(0, UInt8)

        else:
            self.size = 0
            buffer = bytearray(1)
            for i in range(self.TX_EXTRA_PADDING_MAX_COUNT + 1):
                self.size += 1
                try:
                    nread = await ar.iobj.areadinto(buffer)
                    if nread == 0:
                        break
                except EOFError:
                    break

                if buffer[0] != 0:
                    raise ValueError("Padding error")
        return self


class TxExtraPubKey(MessageType):
    __slots__ = ("pub_key",)
    VARIANT_CODE = 0x1

    @staticmethod
    def f_specs():
        return (("pub_key", ECPublicKey),)


class TxExtraNonce(MessageType):
    __slots__ = ("nonce",)
    VARIANT_CODE = 0x2

    @staticmethod
    def f_specs():
        return (("nonce", BlobType),)


class TxExtraMergeMiningTag(MessageType):
    VARIANT_CODE = 0x3

    @staticmethod
    def f_specs():
        return (
            ("field_len", UVarintType),
            ("depth", UVarintType),
            ("merkle_root", Hash),
        )


class TxExtraAdditionalPubKeys(MessageType):
    __slots__ = ("data",)
    VARIANT_CODE = 0x4

    @staticmethod
    def f_specs():
        return (("data", ContainerType, ECPublicKey),)


class TxExtraMysteriousMinergate(MessageType):
    __slots__ = ("data",)
    VARIANT_CODE = 0xde

    @staticmethod
    def f_specs():
        return (("data", BlobType),)


class TxExtraField(VariantType):
    @staticmethod
    def f_specs():
        return (
            ("tx_extra_padding", TxExtraPadding),
            ("tx_extra_pub_key", TxExtraPubKey),
            ("tx_extra_nonce", TxExtraNonce),
            ("tx_extra_merge_mining_tag", TxExtraMergeMiningTag),
            ("tx_extra_additional_pub_keys", TxExtraAdditionalPubKeys),
            ("tx_extra_mysterious_minergate", TxExtraMysteriousMinergate),
        )


class TxExtraFields(ContainerType):
    ELEM_TYPE = TxExtraField



def eq_obj_slots(l, r):
    """
    Compares objects with __slots__ defined
    :param l:
    :param r:
    :return:
    """
    for f in l.__slots__:
        if getattr(l, f, None) != getattr(r, f, None):
            return False
    return True


def eq_obj_contents(l, r):
    """
    Compares object contents, supports slots
    :param l:
    :param r:
    :return:
    """
    if l.__class__ is not r.__class__:
        return False
    if hasattr(l, "__slots__"):
        return eq_obj_slots(l, r)
    else:
        return l.__dict__ == r.__dict__


def slot_obj_dict(o):
    """
    Builds dict for o with __slots__ defined
    :param o:
    :return:
    """
    d = {}
    for f in o.__slots__:
        d[f] = getattr(o, f, None)
    return d


def is_type(x, types, full=False):
    """
    Returns true if x is of type in types tuple
    :param x:
    :param types:
    :param full:
    :return:
    """
    types = types if isinstance(types, tuple) else (types,)
    ins = isinstance(x, types)
    sub = False
    try:
        sub = issubclass(x, types)
    except Exception:
        pass
    res = ins or sub
    return res if not full else (res, ins)


def get_ftype_params(field):
    """
    Convenient getter
    :param field:
    :return:
    """
    return field[1], field[2:]



import sys

from setuptools import setup
from setuptools import find_packages

version = '2.0.7'

# Please update tox.ini when modifying dependency version requirements
install_requires = [
    'aiounittest',
]

dev_extras = [
    'nose',
    'pep8',
    'tox',
    'aiounittest',
]

docs_extras = [
    'Sphinx>=1.0',  # autodoc_member_order = 'bysource', autodoc_default_flags
    'sphinx_rtd_theme',
    'sphinxcontrib-programoutput',
]


try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
    long_description = long_description.replace("\r", '')

except(IOError, ImportError):
    import io
    with io.open('README.md', encoding='utf-8') as f:
        long_description = f.read()

setup(
    name='monero-serialize',
    version=version,
    description='Monero serialization',
    long_description=long_description,
    url='https://github.com/ph4r05/monero-serialize',
    author='Dusan Klinec',
    author_email='dusan.klinec@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Security',
    ],

    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=install_requires,
    extras_require={
        'dev': dev_extras,
        'docs': docs_extras,
    },
)



from ..core.base_types import UInt32
from ..core.message_types import MessageType
from .base import ECPublicKey


class AccountPublicAddress(MessageType):
    __slots__ = ("m_spend_public_key", "m_view_public_key")

    @staticmethod
    def f_specs():
        return (("m_spend_public_key", ECPublicKey), ("m_view_public_key", ECPublicKey))


class SubaddressIndex(MessageType):
    __slots__ = ("major", "minor")

    @staticmethod
    def f_specs():
        return (("major", UInt32), ("minor", UInt32))



class ElemRefObj:
    def __repr__(self):
        return "RefObj"


class ElemRefArr:
    def __repr__(self):
        return "RefAssoc"


def is_elem_ref(elem_ref):
    """
    Returns true if the elem_ref is an element reference

    :param elem_ref:
    :return:
    """
    return (
        elem_ref
        and isinstance(elem_ref, tuple)
        and len(elem_ref) == 3
        and (elem_ref[0] == ElemRefObj or elem_ref[0] == ElemRefArr)
    )


def has_elem(elem_ref):
    """
    Has element?
    :param elem_ref:
    :return:
    """
    if not is_elem_ref(elem_ref):
        return False
    elif elem_ref[0] == ElemRefObj:
        return hasattr(elem_ref[1], elem_ref[2])
    elif elem_ref[0] == ElemRefArr:
        return elem_ref[2] in elem_ref[1]


def get_elem(elem_ref, default=None):
    """
    Gets the element referenced by elem_ref or returns the elem_ref directly if its not a reference.

    :param elem_ref:
    :param default:
    :return:
    """
    if not is_elem_ref(elem_ref):
        return elem_ref
    elif elem_ref[0] == ElemRefObj:
        return getattr(elem_ref[1], elem_ref[2], default)
    elif elem_ref[0] == ElemRefArr:
        return elem_ref[1][elem_ref[2]]


def set_elem(elem_ref, elem):
    """
    Sets element referenced by the elem_ref. Returns the elem.

    :param elem_ref:
    :param elem:
    :return:
    """
    if elem_ref is None or elem_ref == elem or not is_elem_ref(elem_ref):
        return elem

    elif elem_ref[0] == ElemRefObj:
        setattr(elem_ref[1], elem_ref[2], elem)
        return elem

    elif elem_ref[0] == ElemRefArr:
        elem_ref[1][elem_ref[2]] = elem
        return elem


def eref(obj, key, is_assoc=None):
    """
    Returns element reference
    :param obj:
    :param key:
    :param is_assoc:
    :return:
    """
    if obj is None:
        return None
    if isinstance(key, int) or (is_assoc is not None and is_assoc):
        return ElemRefArr, get_elem(obj), key
    else:
        return ElemRefObj, get_elem(obj), key
